from django.db import models 
from choices import StateChoices


# WORKFLOW ITEMS MODEL

class workflowitems(models.Model):
    
    created_date = models.DateTimeField(auto_now_add=True)
    initial_state = models.CharField(max_length=50, default=StateChoices.STATUS_DRAFT)
    interim_state = models.CharField(max_length=50, default=StateChoices.STATUS_DRAFT)
    final_state = models.CharField(max_length=50, default=StateChoices.STATUS_DRAFT)
    # next_available_transitions = ArrayField(models.CharField(max_length=500,blank=True, null=True,default=None),blank=True, null=True,default = None)
    user = models.ForeignKey("accounts.User", on_delete=models.CASCADE, related_name='customername')
    # current_from_party = models.ForeignKey("accounts.Parties", on_delete=models.DO_NOTHING, related_name='from_party')
    # current_to_party = models.ForeignKey("accounts.Parties", on_delete=models.DO_NOTHING, related_name='to_party')
    action = models.CharField(max_length=25, default='SAVE')
    subaction = models.CharField(max_length=55 , blank=True, null=True)
    previous_action = models.CharField(max_length=55 , blank=True, null=True)
    type = models.CharField(max_length=55)
    comments = models.CharField(max_length=500,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)

    class Meta:
        verbose_name_plural = "Workflowitem"


# WORKEVENTS
class workevents(models.Model):

    workitems = models.ForeignKey(workflowitems, on_delete=models.CASCADE, related_name='workflowevent')
    from_state = models.CharField(max_length=50, default='DRAFT')
    action = models.CharField(max_length=25, default='SAVE')
    subaction = models.CharField(max_length=55 , blank=True, null=True)
    to_state = models.CharField(max_length=50, default='DRAFT')
    interim_state = models.CharField(max_length=50, default='DRAFT')
    # from_party = models.ForeignKey('accounts.Parties', on_delete=models.CASCADE, related_name='from_we_party')
    # to_party = models.ForeignKey('accounts.Parties', on_delete=models.CASCADE, related_name='to_wf_party')
    event_user = models.ForeignKey('accounts.User', on_delete=models.CASCADE, related_name='event_user')
    record_datas = models.JSONField(blank=True, null=True)
    end = models.CharField(max_length=55,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    final = models.CharField(max_length=55,blank=True, null=True)
    c_final = models.CharField(max_length=55,blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    created_date = models.DateTimeField(auto_now_add=True)
    type = models.CharField(max_length=55)
    
    class Meta:
        verbose_name_plural = "WorkEvent"
        ordering = ['id']
    
    # def save(self, *args, **kwargs):
    #     try:
    #         qs = Programs.objects.get(id =self.workitems.program.id)
    #         # qs2 = Programs.objects.get(id = self.workitems.invoice.pairing.program_id.id)
    #         # qs1 = Invoices.objects.get(id = self.workitems.invoices)
    #         # qs2 = Invoiceuploads.objects.get(id = self.workitems.uploads)
            
    #         program_data_template = {
    #                 "party" : qs.party.name,
    #                 "party_type" : qs.party.party_type,
    #                 "program_type" : qs.program_type,
    #                 "finance_request_type" : qs.finance_request_type,
    #                 "limit_currency": qs.limit_currency,
    #                 "total_limit_amount":str(qs.total_limit_amount),
    #                 "finance_currency" : qs.finance_currency,
    #                 "settlement_currency" : qs.settlement_currency,
    #                 "expiry_date" : str(qs.expiry_date),
    #                 "max_finance_percentage":str(qs.max_finance_percentage),
    #                 "max_invoice_age_for_funding" : qs.max_invoice_age_for_funding,
    #                 "max_age_for_repayment" : qs.max_age_for_repayment,
    #                 "minimum_period" : qs.minimum_period,
    #                 "maximum_period" : str(qs.maximum_period),
    #                 "maximum_amount" : str(qs.maximum_amount),
    #                 "minimum_amount" : str(qs.minimum_amount),
    #                 "financed_amount":str(qs.financed_amount),
    #                 "balance_amount" : str(qs.balance_amount),
    #                 "grace_period" : qs.grace_period,
    #                 "interest_type" : qs.interest_type.description,
    #                 "interest_rate_type" : qs.interest_rate_type.description,
    #                 "interest_rate": str(qs.interest_rate),
    #                 "margin" : str(qs.margin),
    #                 "comments" : qs.comments,
    #                 "status" : qs.status,
    #                 "is_locked" : qs.is_locked,
    #                 # "created_date" : str(qs.created_date)
    #         }
    #         # invoice_data_template = {
    #         #     "party" : qs1.party.name,
    #         #     "party_type" : qs1.party.party_type,
    #         #     "program_type" : qs1.program_type,
    #         #     "pairing":qs1.pairing,
    #         #     "invoice_no":qs1.invoice_no,
    #         #     "issue_date":qs1.issue_date,
    #         #     "due_date":qs1.due_date,
    #         #     "invoice_currency":qs1.invoice_currency,
    #         #     "amount":qs1.amount,
    #         #     "funding_req_type":qs1.funding_req_type,
    #         #     "finance_currency_type" :qs1.finance_currency_type,
    #         #     "settlement_currency_type" :qs1.settlement_currency_type,
    #         #     "interest_rate" : qs1.interest_rate,
    #         #     "financed_amount" : qs1.financed_amount,
    #         #     "bank_loan_id" : qs1.bank_loan_id,
    #         #     "created_date" : qs1.created_date
    #         # }
    #         self.record_datas = program_data_template
    #         return super(workevents, self).save( *args, **kwargs) 
    #     except:
    #         self.record_datas = None
    #         return super(workevents, self).save( *args, **kwargs)


from asyncore import read
from rest_framework import serializers
from .models import TransitionManager , workevents , workflowitems , Action


class Workeventsserializer(serializers.ModelSerializer):
    event_user = serializers.SlugRelatedField(read_only=True,slug_field = "username")
    class Meta:
        model = workevents
        fields = [
            'id',
            'action',
            'subaction',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'from_party',
            'to_party'
        ]




class Workitemserializer(serializers.ModelSerializer):
    WorkFlowEvents = Workeventsserializer(many=True, read_only=True)
    class Meta:
        model = workflowitems
        fields = [
            'id',
            'transitionmanager',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'current_from_party',
            'current_to_party',
            'WorkFlowEvents',
        ]






class TransitionManagerserializer(serializers.ModelSerializer):
    workflowitems = Workitemserializer(read_only = True)
    class Meta:
        model = TransitionManager
        fields = [
            'id',
            'type',
            'workflowitems'
        ]



class Actionseriaizer(serializers.ModelSerializer):
    class Meta:
        model = Action
        fields = '__all__'



class workflowitemslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workflowitems
        fields = '__all__'


class workeventslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workevents
        fields = '__all__'


from django.db import models
from .middleware import get_current_user
from .validators import validate_postive


try:
   from django.conf import settings
except ImportError:
    raise Exception("settings.py file is required to run this project")



## CORE CONFIGURATIONS CLASSES ##


class Temp(models.Model):
    pass

# Create your models here.

class TransitionManager(models.Model):
    type = models.CharField(max_length = 255 , unique = True)
    sub_sign = models.IntegerField(default = 0 ,  editable = False)

    # default 1 for initial submit and maker process

    def save(self, *args, **kwargs):
        self.type = self.type.upper()
        return super(TransitionManager,self).save(*args, **kwargs)
    
    def __str__(self):
        return self.type

    class Meta:
        verbose_name_plural = "1. TransitionManager"

class States(models.Model):
    description = models.CharField(max_length=255)
    

    def __str__(self):
        return self.description

    class Meta:
        verbose_name_plural = "2. States"


class Action(models.Model):
    description = models.CharField(max_length=255 , blank = True , null=True , help_text = 'e.g., SUBMIT , DELETE') 
    type = models.ForeignKey(TransitionManager , on_delete = models.CASCADE)
    from_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , related_name = 'action_from_state')
    to_state = models.ForeignKey(States , on_delete= models.DO_NOTHING)
    # optional fields
    from_party = models.ForeignKey(settings.FINFLO['FROM_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , help_text = 'this field is optional' , related_name = 'from_transition_party')
    to_party = models.ForeignKey(settings.FINFLO['TO_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , help_text = 'this field is optional' , related_name = 'to_last_party')
    sign_required =  models.IntegerField(default = 0 , validators=[validate_postive] , help_text = "IMPORTANT : if 0 means initial_transition ")

    def save(self, *args, **kwargs):
        self.description = self.description.upper()
        if self.sign_required < 0 :
            self.sign_required = 0
        return super(Action,self).save(*args, **kwargs)

    def __str__(self):
        return "{0} -> {1}".format(self.description , self.type)

    class Meta:
        verbose_name_plural = "3. Action"
        unique_together = ('description', 'type') 
 

class workflowitems(models.Model):
    
    created_date = models.DateTimeField(auto_now_add=True)
    transitionmanager = models.OneToOneField(TransitionManager, on_delete=models.CASCADE,blank=True, null=True )
    initial_state  = models.CharField(max_length=50,default = 'DRAFT')
    interim_state = models.CharField(max_length=50,default = 'DRAFT')
    final_state = models.CharField(max_length=50,default = 'DRAFT')
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    # next_available_transitions = ArrayField(models.CharField(max_length=500,blank=True, null=True,default=None),blank=True, null=True,default = None)
    current_from_party = models.ForeignKey(settings.FINFLO['FROM_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False , related_name = 'wf_from_party')
    current_to_party = models.ForeignKey(settings.FINFLO['TO_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False )
    action = models.CharField(max_length=25 , blank=True, null=True)
    subaction = models.CharField(max_length=55 , blank=True, null=True)
    previous_action = models.CharField(max_length=55 , blank=True, null=True)
    model_type  = models.CharField(max_length=55)
    comments = models.CharField(max_length=500,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)

    class Meta:
        verbose_name_plural = "WorkFlowItem"
        ordering = ['id']

    # def save(self, *args, **kwargs):
    #    if self.from_party is None:
    #         self.from_party = None
    #    super(workflowitems, self).save(*args, **kwargs) # Call the real save() method


# WORKEVENTS
class workevents(models.Model):

    workflowitems = models.ForeignKey(workflowitems, on_delete=models.CASCADE , related_name='WorkFlowEvents')
    action = models.CharField(max_length=25, blank=True, null=True)
    subaction = models.CharField(max_length=55 , blank=True, null=True)
    initial_state  = models.CharField(max_length=50 , default = 'DRAFT')
    interim_state = models.CharField(max_length=50,default = 'DRAFT')
    final_state = models.CharField(max_length=50,default = 'DRAFT')
    from_party = models.ForeignKey(settings.FINFLO['FROM_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False ,related_name = 'we_from_party')
    to_party = models.ForeignKey(settings.FINFLO['TO_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False )
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL,  on_delete=models.CASCADE )
    end_value = models.CharField(max_length=55,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    final_value = models.CharField(max_length=55,blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    created_date = models.DateTimeField(auto_now_add=True)
    type = models.CharField(max_length=55)
    

    class Meta:
        verbose_name_plural = "WorkFlowEvent"
        ordering = ['id']








from django.contrib import admin
from finflo.models import Action, States, Temp, TransitionManager, workflowitems, workevents
# Register your models here.
from django.conf import settings




admin.site.register(TransitionManager)
admin.site.register(Action )
admin.site.register(States)


# sign list fixed  , can be changed dynamically in future



sign_list = [
    "NO_SIGN",
    "AWAITING_SIGN_A",
    "AWAITING_SIGN_B",
    "AWAITING_SIGN_C",
    "AWAITING_SIGN_D",
    "AWAITING_SIGN_E",
    "AWAITING_SIGN_F",
    "AWAITING_SIGN_G",
    "AWAITING_SIGN_H",
    "AWAITING_SIGN_I",
    "AWAITING_SIGN_J",
    "AWAITING_SIGN_K",
    "AWAITING_SIGN_L",
    "AWAITING_SIGN_M",
    "AWAITING_SIGN_N",
    "AWAITING_SIGN_O",
    "AWAITING_SIGN_P",
    "AWAITING_SIGN_Q",
    "AWAITING_SIGN_R",
    "AWAITING_SIGN_S",
    "AWAITING_SIGN_T",
    "AWAITING_SIGN_U",
    "AWAITING_SIGN_V",
    "AWAITING_SIGN_W",
    "AWAITING_SIGN_X",
    "AWAITING_SIGN_Y",
    "AWAITING_SIGN_Z",
]


sub_action = [
    "MAKER",
    "SIGN_A",
    "SIGN_B",
    "SIGN_C",
    "SIGN_D",
    "SIGN_E",
    "SIGN_F",
    "SIGN_G",
    "SIGN_H",
    "SIGN_I",
    "SIGN_J",
    "SIGN_K",
    "SIGN_L",
    "SIGN_M",
    "SIGN_N",
    "SIGN_O",
    "SIGN_P",
    "SIGN_Q",
    "SIGN_R",
    "SIGN_S",
    "SIGN_T",
    "SIGN_U",
    "SIGN_V",
    "SIGN_W",
    "SIGN_X",
    "SIGN_Y",
    "SIGN_Z",
]


from django.http import HttpResponse
from .models import TransitionManager , Action, workevents, workflowitems
from .serializer import TransitionManagerserializer , Actionseriaizer, Workitemserializer, workeventslistserializer, workflowitemslistserializer
from .middleware import get_current_user
from rest_framework.generics import ListAPIView , ListCreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework import status
from .exception import ModelNotfound, SignLengthError, TransitionNotAllowed
from .states import sign_list, sub_action




####################################################
#############       CORE     #######################
####################################################




def gets_wf_item(gets_model):
    ws = workflowitems.objects.get(transitionmanager=gets_model.id)
    return ws


class FinFlotransition:
    
    def delete(self):
        return None

    def transition(self, type, action, stage, id=None):
        try:
            gets_model = TransitionManager.objects.get(type=type.upper())
            gets_action = Action.objects.get(description=action.upper())
        except:
            raise ModelNotfound("no model found ")

        # sign length check
        if id is not None:
            gets_model_id = TransitionManager.objects.get(id=id)
        else:
            print("no transition model found on this id")

        # check sign and stage 
        if stage == gets_model.sub_sign:
            if gets_action.sign_required >= stage:
                def Transition_Handler():
                    gets_sign = gets_action.sign_required

                    # no sign_Required
                    if stage and gets_sign == 0:
                        print("1")
                        gets_model.sub_sign = stage
                        gets_model.save()
                        ws = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model_id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" : gets_action.to_state.description or sign_list[0], 
                            "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sub_action[0], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                        gets_wf = gets_wf_item(gets_model)
                        workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,
                                                  interim_state=gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), final_value="YES" , from_party = gets_action.from_party , to_party = gets_action.to_party)

                    # initial_stage_transition
                    if stage == 0:
                        print("2")
                        gets_model.sub_sign = stage + 1
                        gets_model.save()
                        ws =  workflowitems.objects.update_or_create(transitionmanager=gets_model or gets_model_id, defaults = {"initial_state" : gets_action.from_state.description, "interim_state" : sign_list[1], 
                            "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sub_action[0], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.from_party})
                        gets_wf = gets_wf_item(gets_model)
                        workevents.objects.create(workflowitems=gets_wf or ws, event_user=get_current_user(),  initial_state=gets_action.from_state.description,
                                                  interim_state=sign_list[1], final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper() , from_party = gets_action.from_party , to_party = gets_action.from_party)
                    
                    # final transition
                    elif stage == gets_sign:
                        print("3")
                        gets_model.sub_sign = 0
                        gets_model.save()
                        gets_wf = gets_wf_item(gets_model)
                        workflowitems.objects.filter(id=int(gets_wf.id)).update(
                            initial_state=gets_action.from_state.description, interim_state=gets_action.to_state.description, transitionmanager=gets_model.id or gets_model_id,
                            final_state=gets_action.to_state.description, action=action, subaction=sub_action[int(stage)], model_type=type.upper(), event_user=get_current_user() , current_from_party = gets_action.from_party , current_to_party = gets_action.to_party)
                        
                        workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,
                                                  interim_state=gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[int(stage)], type=type.upper(), final_value="YES" , from_party = gets_action.from_party , to_party = gets_action.to_party)
                    
                    # inbetween all transitions
                    else:
                        print("4")
                        gets_wf = gets_wf_item(gets_model)
                        gets_model.sub_sign = stage + 1
                        gets_model.save()
                        workflowitems.objects.filter(id=int(gets_wf.id)).update(
                            initial_state=gets_action.from_state.description, interim_state=sign_list[
                                1 + stage], transitionmanager=gets_model or gets_model_id,
                            final_state=gets_action.to_state.description, action=action, subaction=sub_action[int(stage)], model_type=type.upper(), event_user=get_current_user(), current_from_party = gets_action.from_party , current_to_party = gets_action.from_party)
                        workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,
                                                  interim_state=sign_list[1 + stage], final_state=gets_action.to_state.description, action=action, subaction=sub_action[int(stage)], type=type.upper() , from_party = gets_action.from_party , to_party = gets_action.from_party)

                return Transition_Handler()
            else:
                raise SignLengthError(
                    "either the stage nor the sign_required length mismatching and the stage should not be zero ")
        else:
            raise TransitionNotAllowed(
                "TransitionNotAllowed please try again")






####################################################
#############       API      #######################
####################################################



class DetailsListApiView(ListAPIView):
    queryset = TransitionManager.objects.all()
    serializer_class = TransitionManagerserializer
    permission_classes = [IsAuthenticated]

    # def get_queryset(self, request):
    #     type = self.request.query_params.get('type')
    #     if type is None:
    #         queryset = TransitionManager.objects.all()
    #     queryset = TransitionManager.objects.filter(type=type)
    #     return queryset

    def list(self, request):
        queryset = TransitionManager.objects.all()
        serializer = TransitionManagerserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)



# worflow api 

class WorkFlowitemsListApi(ListAPIView):
    queryset = workflowitems.objects.all()
    serializer_class = Workitemserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = workflowitems.objects.all()
        serializer = Workitemserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)




# workevents api 


class WorkEventsListApi(ListAPIView):
    queryset = workevents.objects.all()
    serializer_class = workeventslistserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = workevents.objects.all()
        serializer = workeventslistserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)






# action create and list api 


class ActionListApi(ListCreateAPIView):
    queryset = Action.objects.all()
    serializer_class = Actionseriaizer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = Action.objects.all()
        serializer = Actionseriaizer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)

    def post(self, request):
        serializer = Actionseriaizer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)









## BASIC EXCEPTION CLASSES 


class ActionNotFound(Exception):
    pass


class TypeDoesNotExist(Exception):
    pass

class TypeEmpty(Exception):
    pass

class ModelNotfound(Exception):
    pass

class MoreThanOneModel(Exception):
    pass

class IDError(Exception):
    pass

class SignLengthError(Exception):
    pass


class TransitionNotAllowed(Exception):
    pass


from rest_framework.views import APIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from .transition import FinFlotransition

class TransitionApiView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self,request):
        qs = FinFlotransition()
        type = request.GET.get("type","")
        action = request.GET.get("action","")
        stage = request.GET.get("stage","")
        t_id = request.GET.get("t_id","")
        if (type , action , stage ) is not None:
            qs.transition(type = type , action = action , stage = stage , id = t_id)
            return Response({"status" : "Transition success"},status = status.HTTP_200_OK)
        return Response({"status" : "Transition failure"},status = status.HTTP_204_NO_CONTENT)

